source("R/BevertonHolt_models.R")
source("R/SEbootstrap_caracoles.R")
source("R/cxr_optimize.R")
source("R/InitParams.R")
source("R/RetrieveParams.R")
source("R/GenerateTestData.R")
# optimization methods
library(nloptr)
library(GenSA)
library(hydroPSO)
library(DEoptimR)
#other packages
library(tidyverse)

###########################
# test data
focal.sp <- c(1,2)
num.sp <- 5
num.cov <- 2
num.obs <- 25 # per focal species

focal.lambda <- c(10,20)
alpha.matrix.orig <- matrix(data = runif(num.sp*num.sp,-0.001,0),nrow = num.sp, ncol = num.sp)
alpha.cov.orig <- list()
lambda.cov.orig <- list()

for(i.sp in 1:num.sp){
  alpha.cov.orig[[i.sp]] <- matrix(data = rnorm(num.sp*num.cov,0,0.005),nrow = num.sp, ncol = num.cov) # rows: competitors, columns: covariates
  lambda.cov.orig[[i.sp]] <- rnorm(num.cov,0,0.005)#rep(0,num.cov)
}

competition.data <- GenerateTestData(focal.sp = focal.sp,
                              num.sp = num.sp,
                              num.cov = num.cov,
                              num.obs = num.obs,
                              fitness.model = 5,
                              focal.lambda = focal.lambda,
                              alpha.matrix = alpha.matrix.orig,
                              alpha.cov = alpha.cov.orig,
                              lambda.cov = lambda.cov.orig)
comp.matrix <- competition.data[,2:(num.sp+1)]
covariates <- competition.data[,(num.sp+2):(num.sp+2+num.cov-1)]

#############################
# simulation parameters

# models to parameterize
# be aware of including 4 and 5 ONLY if there are covariates
# otherwise it makes no sense (see equations in Lanuza et al. 2018)
models <- 1:5

# which values do we optimize for each model?
param.list <- list(c("lambda"),
                   c("lambda","alpha"),
                   c("lambda","alpha"),
                   c("lambda","alpha","lambda.cov","alpha.cov"),
                   c("lambda","alpha","lambda.cov","alpha.cov"))

# keep the model definitions in a list, for ease
fitness.models <- list(BH_1 = BH_1,BH_2 = BH_2,BH_3 = BH_3,BH_4 = BH_4,BH_5 = BH_5)

# optimization methods to use
optim.methods <- c("optim_NM"#,
                   # "optim_L-BFGS-B",
                   # "nloptr_CRS2_LM", 
                   # "nloptr_ISRES", 
                   # "nloptr_DIRECT_L_RAND", 
                   # "GenSA", 
                   # "hydroPSO", 
                   # "DEoptimR"
)

# from which method are we taking initial estimates for the next model?
# preliminary observations suggest optim_NM or DEoptimR
init.par.method <- optim.methods[1]
init.method.num <- which(optim.methods == init.par.method)

# values for initial estimates of parameters. 
# Overwrite if necessary
# init.lambda is calculated after log.fitness for each focal species
lower.lambda <- 1
upper.lambda <- 1e4
# sigma
lower.sigma <- 0
upper.sigma <- 1
# alpha
init.alpha <- 1e-4
lower.alpha <- 0
upper.alpha <- 1e5
# lambda.cov
init.lambda.cov <- 1
lower.lambda.cov <- 1e-4
upper.lambda.cov <- 1e4
# alpha.cov
init.alpha.cov <- 1
lower.alpha.cov <- 1e-4
upper.alpha.cov <- 1e4

# if we want quick calculations, we can disable 
# the bootstrapping for the standard errors
generate.errors <- TRUE
bootstrap.samples <- 5

# store results?
write.results <- FALSE

##############################
# initialize data structures
# elements are matrices/vectors within nested lists, of the form matrix[[focal.sp]][[model]][[method]]

param.matrices <- list()
for(i.sp in 1:length(focal.sp)){
  param.matrices[[i.sp]] <- list()
  for(i.model in 1:length(models)){
    param.matrices[[i.sp]][[i.model]] <- list()
    for(i.method in 1:length(optim.methods)){
      param.matrices[[i.sp]][[i.model]][[i.method]] <- list(lambda = 0,
                                                            lambda.lower.error = 0,
                                                            lambda.upper.error = 0,
                                                            sigma = 0,
                                                            alpha = 0,
                                                            alpha.lower.error = 0,
                                                            alpha.upper.error = 0,
                                                            lambda.cov = 0,
                                                            lambda.cov.lower.error = 0,
                                                            lambda.cov.upper.error = 0,
                                                            alpha.cov = 0,
                                                            alpha.cov.lower.error = 0,
                                                            alpha.cov.upper.error = 0,
                                                            log.likelihood = 0)
    }
    names(param.matrices[[i.sp]][[i.model]]) <- optim.methods
  }
  names(param.matrices[[i.sp]]) <- names(fitness.models)[models]
}
names(param.matrices) <- focal.sp

###############################
# main loop

for(i.sp in 1:length(focal.sp)){
  
  # subset and prepare the data
  
  focal.sp.data <- subset(competition.data, focal == focal.sp[i.sp])
  # current focal species
  focal <- unique(focal.sp.data$focal)
  # fitness metric...
  # subset >0 records, for calculating logarithms
  focal.sp.data <- subset(focal.sp.data, fitness > 0)
  fitness <- focal.sp.data$fitness #fitness
  log.fitness <- log(fitness)
  # competition matrix: number of competitors
  focal.comp.matrix <- comp.matrix[which(competition.data$focal == focal.sp[i.sp]),]
  # number of competitors
  num.competitors <- dim(focal.comp.matrix)[2]
  # number of covariates
  num.covariates <- ifelse(is.null(ncol(covariates)),0,ncol(covariates))
  # covariates for the focal species
  if(num.covariates > 0){
    focal.covariates <- covariates[which(competition.data$focal == focal.sp[i.sp]),,drop = FALSE]
  }else{
    focal.covariates <- 0
  }
  
  # model to optimize  
  for(i.model in 1:length(models)){
    
    print("*********************************")
    print(paste(date()," - starting focal sp ",focal.sp[i.sp],", model ",models[i.model],sep=""))
    print("*********************************")
    #message()
    
    # obtain initial estimates from either previous model, or from given values
    # also, beware if the initial estimates are single values or vectors.
    
    # lambda
    if("lambda" %in% param.list[[i.model]]){
      if(i.model != 1){
        my.init.lambda <- param.matrices[[i.sp]][[i.model-1]][[init.par.method]]$lambda
      }else{
        my.init.lambda <- mean(log.fitness)
      }
    }else{
      my.init.lambda <- init.lambda[i.sp]
    }
    
    # sigma
    if(i.model != 1){
      my.init.sigma <- param.matrices[[i.sp]][[i.model-1]][[init.par.method]]$sigma
    }else{
      my.init.sigma <- sd(log.fitness)
    }
    
    # alpha
    if("alpha" %in% param.list[[i.model]]){
      
      # if first model with alpha
      if(models[i.model] <= 2){
        my.init.alpha <- init.alpha
      }else{
        # if previous estimate
        if(i.model > 1){
          # if length of previous estimate is 1, expand it
          if(length(param.matrices[[i.sp]][[i.model-1]][[init.par.method]]$alpha) == 1){
            my.init.alpha <- rep(length(param.matrices[[i.sp]][[i.model-1]][[init.par.method]]$alpha) == 1,num.competitors)
          }else{
            my.init.alpha <- param.matrices[[i.sp]][[i.model-1]][[init.par.method]]$alpha
          }
          # if no previous estimate, expand the initial value
        }else{
          my.init.alpha <- rep(init.alpha,num.competitors)
        }
      }
    }else{
      my.init.alpha <- init.alpha[i.sp]
    }
    
    # lambda.cov
    if("lambda.cov" %in% param.list[[i.model]]){
      
      if(models[i.model] == 4){
        my.init.lambda.cov <- rep(init.lambda.cov,num.covariates)
      }else if(models[i.model] == 5){
        if(i.model == 1){
          my.init.lambda.cov <- rep(init.lambda.cov,num.covariates)
        }else{
          my.init.lambda.cov <- param.matrices[[i.sp]][[i.model-1]][[init.par.method]]$lambda.cov
        }
      }
    }else{
      my.init.lambda.cov <- init.lambda.cov[i.sp]
    }
    
    # alpha.cov
    if("alpha.cov" %in% param.list[[i.model]]){
      if(models[i.model] == 4){
        my.init.alpha.cov <- rep(init.alpha.cov,num.covariates)
      }else if(models[i.model] == 5){
        if(i.model > 1){
          my.init.alpha.cov <- rep(param.matrices[[i.sp]][[i.model-1]][[init.par.method]]$alpha.cov,each=num.competitors)
        }else{
          my.init.alpha.cov <- rep(init.alpha.cov,num.competitors*num.covariates)
        }
      }
    }else{
      my.init.alpha.cov <- init.alpha.cov[i.sp]
    }
    
    ######################
    # compute each method
    
    for(i.method in 1:length(optim.methods)){
      
      temp.results <- cxr_optimize(fitness.model = fitness.models[[models[i.model]]],
                                   optim.method = optim.methods[i.method],
                                   param.list = param.list[[i.model]],
                                   log.fitness = log.fitness,
                                   init.lambda = my.init.lambda,
                                   lower.lambda = lower.lambda,
                                   upper.lambda = upper.lambda,
                                   init.sigma = my.init.sigma,
                                   lower.sigma = lower.sigma,
                                   upper.sigma = upper.sigma,
                                   init.alpha = my.init.alpha,
                                   lower.alpha = lower.alpha,
                                   upper.alpha = upper.alpha,
                                   init.lambda.cov = my.init.lambda.cov,
                                   lower.lambda.cov = lower.lambda.cov,
                                   upper.lambda.cov = upper.lambda.cov,
                                   init.alpha.cov = my.init.alpha.cov,
                                   lower.alpha.cov = lower.alpha.cov,
                                   upper.alpha.cov = upper.alpha.cov,
                                   focal.comp.matrix = focal.comp.matrix,
                                   focal.covariates = focal.covariates,
                                   generate.errors = generate.errors,
                                   bootstrap.samples = bootstrap.samples)
      ###############
      # clean up results
      
      param.matrices[[i.sp]][[i.model]][[i.method]]$lambda <- temp.results$lambda
      param.matrices[[i.sp]][[i.model]][[i.method]]$lambda.lower.error <- temp.results$lambda.lower.error
      param.matrices[[i.sp]][[i.model]][[i.method]]$lambda.upper.error <- temp.results$lambda.upper.error
      
      param.matrices[[i.sp]][[i.model]][[i.method]]$sigma <- temp.results$sigma
      
      param.matrices[[i.sp]][[i.model]][[i.method]]$alpha <- temp.results$alpha
      param.matrices[[i.sp]][[i.model]][[i.method]]$alpha.upper.error <- temp.results$alpha.upper.error
      param.matrices[[i.sp]][[i.model]][[i.method]]$alpha.lower.error <- temp.results$alpha.lower.error
      
      param.matrices[[i.sp]][[i.model]][[i.method]]$lambda.cov <- temp.results$lambda.cov
      param.matrices[[i.sp]][[i.model]][[i.method]]$lambda.cov.upper.error <- temp.results$lambda.cov.upper.error
      param.matrices[[i.sp]][[i.model]][[i.method]]$lambda.cov.lower.error <- temp.results$lambda.cov.lower.error
      
      param.matrices[[i.sp]][[i.model]][[i.method]]$alpha.cov <- temp.results$alpha.cov
      param.matrices[[i.sp]][[i.model]][[i.method]]$alpha.cov.upper.error <- temp.results$alpha.cov.upper.error
      param.matrices[[i.sp]][[i.model]][[i.method]]$alpha.cov.lower.error <- temp.results$alpha.cov.lower.error
      
      param.matrices[[i.sp]][[i.model]][[i.method]]$log.likelihood <- temp.results$log.likelihood
      
    }# for i.method
  }# for i.model
}# for i.sp

if(write.results){
  save(param.matrices,file = "./results/param_estimates.Rdata")
}

